/*
 * VeriSilicon VPE Video Encoder Common Interface
 * Copyright (C) 2020 VeriSilicon Holdings Co., Ltd.
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <stdio.h>
#include <stddef.h>

#include "vpi_api.h"
#include "vpi_types.h"

#include "codec_internal.h"
#include "internal.h"
#include "hwconfig.h"
#include "encode.h"
#include "libavutil/hwcontext_vpe.h"
#include "libavutil/time.h"
#include "libavutil/opt.h"
#include "libavutil/pixfmt.h"
#include "libavutil/pixdesc.h"

typedef struct VpeEncCtx {
    const AVClass *class;

    AVBufferRef *hwframe;
    AVBufferRef *hwdevice;


    VpiEncHandle enc_handle;
    VpiEncInitParams *init_params;

    int eof;

    AVFrame *frame;
    /*Encoding preset, superfast/fast/medium/slow/superslow*/
    char *preset;
    /*Profile of the encoding*/
    char *profile;
    /*Level of the encoding*/
    char *level;
    /*Encoding parameters*/
    char *enc_params;
    /*VCE Constant rate factor mode*/
    int crf;
    /*Force IDR*/
    char *force_idr;
    /*Force 8bit output*/
    int force8;

    void *input_frm_pool;

    VpiPkt *vpi_pkt;

    int initialized;
} VpeEncCtx;

#define FRAME_BUF_POOL_SIZE   128
static int vpe_encode_start(AVCodecContext *avctx);
static int vpe_enc_init_hwctx(AVCodecContext *avctx)
{
    VpeEncCtx *ctx = (VpeEncCtx *)avctx->priv_data;
    AVVpeDeviceContext *device_hwctx;
    AVVpeFramesContext *frame_hwctx;
    AVHWFramesContext *hwframe_ctx;
    AVHWDeviceContext *hwdevice_ctx;
    AVVpeHWConfig *hwconfig = NULL;
    AVHWFramesConstraints *constraints = NULL;
    int ret = 0;
    int i;

    /*Get HW frame. The avctx->hw_frames_ctx is the reference to the
      AVHWFramesContext describing the input frame for vpe encoder*/
    if (avctx->hw_frames_ctx) {
        ctx->hwframe = av_buffer_ref(avctx->hw_frames_ctx);
        if (!ctx->hwframe) {
            av_log(avctx, AV_LOG_DEBUG, "hwframe ref failed\n");
            return AVERROR(ENOMEM);
        }
    } else {
        av_log(avctx, AV_LOG_DEBUG, "hw_frames_ctx NULL\n");
        return AVERROR_INVALIDDATA;
    }
    hwframe_ctx = (AVHWFramesContext*)ctx->hwframe->data;
    if (avctx->hw_device_ctx) {
        hwconfig = av_hwdevice_hwconfig_alloc(avctx->hw_device_ctx);
    } else {
        ctx->hwdevice = av_buffer_ref(hwframe_ctx->device_ref);
        if (!ctx->hwdevice) {
            av_log(avctx, AV_LOG_DEBUG, "hwdevice ref failed\n");
            return AVERROR(ENOMEM);
        }
        hwconfig = av_hwdevice_hwconfig_alloc(ctx->hwdevice);
    }
    if (!hwconfig) {
        av_log(avctx, AV_LOG_DEBUG, "hwconfig NULL\n");
        return AVERROR(ENOMEM);
    }

    switch (avctx->codec_id) {
    case AV_CODEC_ID_H264:
        hwconfig->id  = VPE_HW_H264ENC;
        break;
    case AV_CODEC_ID_HEVC:
        hwconfig->id  = VPE_HW_HEVCENC;
        break;
    case AV_CODEC_ID_AV1:
        if (!strcmp(avctx->codec->name, "xav1enc_vpe")) {
            hwconfig->id  = VPE_HW_XAV1ENC;
        } else {
            hwconfig->id  = VPE_HW_AV1ENC;
        }
        break;
    case AV_CODEC_ID_MJPEG:
        hwconfig->id  = VPE_HW_JPEGENC;
        break;
    default:
        ret = AVERROR(EINVAL);
        goto error_exit;
    }
    av_log(avctx, AV_LOG_DEBUG, "hwconfig->id = %d\n", hwconfig->id);

    if (avctx->hw_device_ctx) {
        constraints =
            av_hwdevice_get_hwframe_constraints(avctx->hw_device_ctx, hwconfig);
    } else {
        constraints =
            av_hwdevice_get_hwframe_constraints(ctx->hwdevice, hwconfig);
    }
    if (!constraints) {
        ret = AVERROR(ENOMEM);
        goto error_exit;
    }

    av_log(avctx, AV_LOG_DEBUG, "get constraints %d,%d,%d,%d\n", constraints->min_width,
                    constraints->min_height, constraints->max_width, constraints->max_height);

    if (avctx->coded_width  < constraints->min_width  ||
        avctx->coded_height < constraints->min_height ||
        avctx->coded_width  > constraints->max_width  ||
        avctx->coded_height > constraints->max_height) {
        av_log(avctx, AV_LOG_ERROR, "Hardware does not support image "
               "size %dx%d (constraints: width %d-%d height %d-%d).\n",
               avctx->coded_width, avctx->coded_height,
               constraints->min_width,  constraints->max_width,
               constraints->min_height, constraints->max_height);
        ret = AVERROR(EINVAL);
        goto error_exit;
    }

    if (!constraints->valid_sw_formats ||
        constraints->valid_sw_formats[0] == AV_PIX_FMT_NONE) {
        av_log(avctx, AV_LOG_ERROR, "Hardware does not offer any "
               "usable surface formats.\n");
        ret = AVERROR(EINVAL);
        goto error_exit;
    }

    av_log(avctx, AV_LOG_DEBUG, "avctx->sw_pix_fmt %s\n",
                                 av_get_pix_fmt_name(avctx->sw_pix_fmt));
    av_log(avctx, AV_LOG_DEBUG, "hwframe_ctx->sw_format %s\n",
                                 av_get_pix_fmt_name(hwframe_ctx->sw_format));

    i = 0;
    while (1) {
        if (avctx->sw_pix_fmt == constraints->valid_sw_formats[i])
            break;
        if (constraints->valid_sw_formats[i] == AV_PIX_FMT_NONE) {
            av_log(ctx, AV_LOG_ERROR, "Hardware can't find any matched sw format\n");
            ret = AVERROR(EINVAL);
            goto error_exit;
        }
        i++;
    }
    av_log(avctx, AV_LOG_DEBUG, "get sw constraints %s\n",
                    av_get_pix_fmt_name(constraints->valid_sw_formats[i]));

    if (!constraints->valid_hw_formats ||
         constraints->valid_hw_formats[0] != AV_PIX_FMT_VPE) {
        av_log(avctx, AV_LOG_ERROR, "Hardware error format\n");
        ret = AVERROR(EINVAL);
        goto error_exit;
    }
    av_log(avctx, AV_LOG_DEBUG, "get hw constraints %s\n",
                    av_get_pix_fmt_name(constraints->valid_hw_formats[0]));

    if (!avctx->hw_device_ctx) {
        device_hwctx = hwframe_ctx->device_ctx->hwctx;
    } else {
        hwdevice_ctx = (AVHWDeviceContext *)avctx->hw_device_ctx->data;
        device_hwctx = (AVVpeDeviceContext *)hwdevice_ctx->hwctx;
    }
    frame_hwctx  = hwframe_ctx->hwctx;

    ctx->init_params->dev_ctx = device_hwctx->vpe_dev_ctx;
    ctx->init_params->mem_ctx = frame_hwctx->vpe_mem_ctx;

    av_log(avctx, AV_LOG_DEBUG, "coded param: %d x %d\n", avctx->coded_width, avctx->coded_height);
    av_log(avctx, AV_LOG_DEBUG, "hwframe_ctx->format = %d\n", hwframe_ctx->format);
    av_log(avctx, AV_LOG_DEBUG, "hwframe_ctx->sw_format = %d\n", hwframe_ctx->sw_format);

    ctx->init_params->orig_width   = avctx->width;
    ctx->init_params->orig_height  = avctx->height;
    ctx->init_params->width        = avctx->width;
    ctx->init_params->height       = avctx->height;

    if (hwframe_ctx->sw_format == AV_PIX_FMT_P010LE ||
        hwframe_ctx->sw_format == AV_PIX_FMT_YUV420P10LE) {
        ctx->init_params->bitdepth = 10;
    } else {
        ctx->init_params->bitdepth = 8;
    }

    av_hwframe_constraints_free(&constraints);
    av_freep(&hwconfig);

    return 0;

error_exit:

    if (constraints) {
        av_hwframe_constraints_free(&constraints);
    }
    if (hwconfig) {
        av_freep(&hwconfig);
    }

    return ret;
}

/**
 * Init vpe enc hw ctx and create vpe context
 */
static av_cold int vpe_encode_init(AVCodecContext *avctx)
{
    VpeEncCtx *ctx = (VpeEncCtx *)avctx->priv_data;
    int ret = 0;

    ctx->init_params = (VpiEncInitParams *)vpi_memory_get_struct(VPE_ENC_INIT_PARAM);
    if (ctx->init_params == NULL) {
        av_log(avctx, AV_LOG_ERROR, "vpi_memory_get_struct enc_init_param failed\n");
        return AVERROR(ENOMEM);
    }
    ret = vpe_enc_init_hwctx(avctx);
    if (ret) {
        av_log(avctx, AV_LOG_ERROR, "enc init hwctx failed\n");
        return ret;
    }
    ctx->init_params->slice_id = -1;

    ret = vpi_pool_create(ctx->init_params->mem_ctx, FRAME_BUF_POOL_SIZE,
                          &ctx->input_frm_pool);
    if (ret < 0) {
        av_log(avctx, AV_LOG_ERROR, "vpi create input frm pool failed %s\n", vpi_error_str(ret));
        return AVERROR(ENOMEM);
    }

    ctx->init_params->enc_type     = VPE_ENCODER_VC8000E;
    ctx->init_params->rate_num     = avctx->framerate.num;
    ctx->init_params->rate_denom   = avctx->framerate.den;
    ctx->init_params->preset       = ctx->preset;
    ctx->init_params->crf          = ctx->crf;
    ctx->init_params->profile      = ctx->profile;
    ctx->init_params->level        = ctx->level;
    ctx->init_params->force_idr    = ctx->force_idr;
    ctx->init_params->force_8bit   = ctx->force8;
    ctx->init_params->bit_rate     = avctx->bit_rate;
    ctx->init_params->input_frm_pool = ctx->input_frm_pool;
    if (!strcmp(avctx->codec->name, "h264enc_vpe")) {
        ctx->init_params->codec_format = VPE_ENC_H264;
    } else if (!strcmp(avctx->codec->name, "hevcenc_vpe")) {
        ctx->init_params->codec_format = VPE_ENC_HEVC;
    } else if (!strcmp(avctx->codec->name, "av1enc_vpe")) {
        ctx->init_params->codec_format = VPE_ENC_AV1;
    } else if (!strcmp(avctx->codec->name, "jpegenc_vpe")) {
        ctx->init_params->codec_format = VPE_ENC_JPEG;
	} else if (!strcmp(avctx->codec->name, "xav1enc_vpe")) {
        ctx->init_params->codec_format = VPE_ENC_AV1;
        ctx->init_params->enc_type     = VPE_ENCODER_XAV1;
        ctx->init_params->xav1_options = ctx->enc_params;
    } else {
        av_log(avctx, AV_LOG_ERROR, "not support %s, please check your params!\n", avctx->codec->name);
        return AVERROR(EINVAL);
    }

    ctx->frame = av_frame_alloc();
    if (!ctx->frame)
        return AVERROR(ENOMEM);

    ctx->vpi_pkt = vpi_memory_get_struct(VPE_PACKET);
    if (!ctx->vpi_pkt)
        return AVERROR(ENOMEM);

    ret = vpe_encode_start(avctx);
    if (ret) {
        av_log(ctx, AV_LOG_ERROR, "vpe_encode_start failed\n");
        return ret;
    }
    return 0;
}

static int vpe_encode_start(AVCodecContext *avctx)
{
    VpeEncCtx *ctx = (VpeEncCtx *)avctx->priv_data;
    AVHWFramesContext *hwframe_ctx;
    AVVpeFramesContext *frame_hwctx;
    VpiFrm *frame;
    unsigned char *extradata;
    int extradata_size;
    int out_index, ret;
    int ch_format;
    char en_flag[2];
    int pp_index;
    AVFrame *first_pic;

    first_pic = (AVFrame *)avctx->opaque;
    if(!first_pic){
        av_log(avctx, AV_LOG_ERROR, "empty first_pic\n");
        return AVERROR_EXTERNAL;
    }

    hwframe_ctx = (AVHWFramesContext*)ctx->hwframe->data;
    frame_hwctx = hwframe_ctx->hwctx;
    frame = (VpiFrm *)first_pic->data[0];

    ret = vpi_memory_get_channel_info(frame_hwctx->vpe_mem_ctx, &frame_hwctx->proc_info->ch_info[0]);
    if (ret < 0) {
        av_log(avctx, AV_LOG_ERROR, "vpi_memory_get_info failed, error=%s\n", vpi_error_str(ret));
        return AVERROR_EXTERNAL;
    }

    if (first_pic->buf[1]) {
        out_index = *(int *)(first_pic->buf[1]->data);
    } else {
        out_index = 0;
    }

    ret = vpi_memory_get_map_info(frame_hwctx->vpe_mem_ctx, out_index, &pp_index);
    if (ret < 0) {
        av_log(avctx, AV_LOG_ERROR, "vpi_memory_get_map_info failed, error=%s\n",
                                    vpi_error_str(ret));
        return AVERROR_EXTERNAL;
    }
    ctx->init_params->enc_index = out_index;
    ctx->init_params->ch_index = pp_index;

    ctx->init_params->vui_param.vui_video_signal_type_present_flag = frame->vui_param.vui_video_signal_type_present_flag;
    if(ctx->init_params->vui_param.vui_video_signal_type_present_flag){
        ctx->init_params->vui_param.vui_video_format = frame->vui_param.vui_video_format;
        ctx->init_params->vui_param.vui_video_full_range = frame->vui_param.vui_video_full_range;
        ctx->init_params->vui_param.color_description_present_flag = frame->vui_param.color_description_present_flag;
        if(ctx->init_params->vui_param.color_description_present_flag){
            memcpy(&ctx->init_params->vui_param.color_description, &frame->vui_param.color_description, sizeof(VpiVuiColorDescription));
        }
    }

    av_log(avctx, AV_LOG_DEBUG,
           "frame_hwctx->proc_info->ch_info[%d].flag = %d\n",
           pp_index, frame_hwctx->proc_info->ch_info[pp_index].flag);
    av_log(avctx, AV_LOG_DEBUG,
           "frame_hwctx->proc_info->ch_info[%d].format = %d\n",
           pp_index, frame_hwctx->proc_info->ch_info[pp_index].format);
    av_log(avctx, AV_LOG_DEBUG,
           "frame_hwctx->proc_info->ch_info[%d].sw_format = %d\n",
           pp_index, frame_hwctx->proc_info->ch_info[pp_index].sw_format);

    if (!strcmp(avctx->codec->name, "xav1enc_vpe")) {
        ctx->init_params->enc_index = 0;
        ctx->init_params->walign   = frame_hwctx->proc_info->ch_info[pp_index].walign;
        ctx->init_params->halign   = frame_hwctx->proc_info->ch_info[pp_index].halign;
        ctx->init_params->flag     = frame_hwctx->proc_info->ch_info[pp_index].flag;
    }

    ch_format = frame_hwctx->proc_info->ch_info[pp_index].format;
    if (ch_format == VPI_FORMAT_VPE) {
        ctx->init_params->input_format = frame_hwctx->proc_info->ch_info[pp_index].sw_format;
    } else {
        av_log(avctx, AV_LOG_ERROR, "input format not support, please check your params!\n");
        return AVERROR_INVALIDDATA;
    }

    if (ctx->init_params->input_format == VPI_FORMAT_P010LE ||
        ctx->init_params->input_format == VPI_FORMAT_YUV420P10LE ||
        ctx->init_params->input_format == VPI_FORMAT_SP1010) {
        if(frame->hdr_param){
            if(frame->hdr_param->mdcv.present_flag){
                memcpy(&ctx->init_params->mdcv, &frame->hdr_param->mdcv, sizeof(VpiMasterDisColVol));
            }
            if(frame->hdr_param->cll.present_flag){
                memcpy(&ctx->init_params->cll, &frame->hdr_param->cll, sizeof(VpiContentLightLevel));
            }
            if(frame->hdr_param->atc.present_flag){
                memcpy(&ctx->init_params->atc, &frame->hdr_param->atc, sizeof(VpiATCInfo));
            }
        }
    }

    av_log(avctx, AV_LOG_DEBUG,
           "w x h = %d x %d, hwframe_ctx->sw_format = %d, format is %d, "
           "ch_format = %d\n",
           ctx->init_params->width, ctx->init_params->height,
           hwframe_ctx->sw_format, ctx->init_params->input_format, ch_format);
    av_log(avctx, AV_LOG_DEBUG, "ch_w x ch_h = %d x %d\n",
           frame_hwctx->proc_info->ch_info[pp_index].width,
           frame_hwctx->proc_info->ch_info[pp_index].height);

    av_log(avctx, AV_LOG_DEBUG, "input_format = %d\n", ctx->init_params->input_format);

    ctx->enc_handle = vpi_encoder_init(ctx->init_params);
    if (ctx->enc_handle == NULL) {
        av_log(avctx, AV_LOG_ERROR, "vpi_encoder_init failed, error=%s\n",
                                     vpi_error_str(ret));
        return AVERROR_EXTERNAL;
    }

    if (ctx->force8) {
        strcpy(en_flag, "1");
        if (vpi_encoder_set_param(ctx->enc_handle, "force8", en_flag) < 0) {
            av_log(avctx, AV_LOG_WARNING,
                   "Error parsing option '%s = %s'. Please check if it is valid for vpe encoder options.\n",
                   "force8", "1");
        }
    }

    /* enc_params get */
    if (ctx->enc_params) {
        AVDictionary *dict    = NULL;
        AVDictionaryEntry *en = NULL;

        if (!av_dict_parse_string(&dict, ctx->enc_params, "=", ":", 0)) {
            while ((en = av_dict_get(dict, "", en, AV_DICT_IGNORE_SUFFIX))) {
                if (vpi_encoder_set_param(ctx->enc_handle,
                                          en->key, en->value) < 0)
                    av_log(avctx, AV_LOG_WARNING,
                           "Error parsing option '%s = %s'. Please check if it is valid enc_params option.\n",
                            en->key, en->value);
            }
            av_dict_free(&dict);
        }
    }

    ret = vpi_encoder_open(ctx->enc_handle);
    if (ret) {
        av_log(avctx, AV_LOG_ERROR, "vpi_encoder_open failed, error=%s\n",
                                     vpi_error_str(ret));
        return AVERROR_EXTERNAL;
    }

    extradata = vpi_encoder_get_extradata(ctx->enc_handle, &extradata_size);
    if (extradata) {
        avctx->extradata = av_malloc(extradata_size);
        memcpy(avctx->extradata, extradata, extradata_size);
        avctx->extradata_size = extradata_size;
    }

    avctx->opaque = NULL;

    ctx->initialized = 1;

    return 0;
}

/**
 * Fetch new AVFrame and put to external encoder
 */
static int vpe_enc_receive_pic(AVCodecContext *avctx)
{
    int ret              = 0;
    VpeEncCtx *ctx       = (VpeEncCtx *)avctx->priv_data;
    VpiFrm *vpi_frame    = NULL;
    AVFrame *frame       = ctx->frame;
    AVFrame *trans_frame = NULL;

    if (!frame->buf[0]) {
        ret = ff_encode_get_frame(avctx, frame);
        if (ret < 0 && ret != AVERROR_EOF)
            return ret;
        if (ret == AVERROR_EOF)
            frame = NULL;
    } else {
        av_log(ctx, AV_LOG_ERROR, "frame has not been put to VPE encoder\n");
        return AVERROR_INVALIDDATA;
    }

    if (frame) {
        trans_frame = av_frame_alloc();
        if (!trans_frame)
            return AVERROR(ENOMEM);
        av_frame_move_ref(trans_frame, frame);
        vpi_frame = (VpiFrm *)trans_frame->data[0];
        vpi_frame->pts = trans_frame->pts;
        vpi_frame->dts = trans_frame->pkt_dts;
        vpi_frame->opaque = (void *)trans_frame;
    } else {
        av_log(ctx, AV_LOG_DEBUG, "input image is empty, received EOF\n");
        ctx->eof = 1;
    }

    ret = vpi_encoder_put_frame(ctx->enc_handle, vpi_frame);
    if (ret) {
        av_log(ctx, AV_LOG_ERROR, "put frame to external encoder failed, error=%s\n",
                                       vpi_error_str(ret));
        return AVERROR_EXTERNAL;
    }

    return 0;
}

/**
 * Translate VpiPacket to AVPacket
 */
static void vpe_encode_output_packet(VpiPkt *vpi_packet,
                                     AVPacket *out_packet)
{
    out_packet->size  = vpi_packet->size;
    out_packet->pts   = vpi_packet->pts;
    out_packet->dts   = vpi_packet->dts;
}

/**
 * Get encoded packet from external encoder
 */
static int vpe_encode_receive_packet(AVCodecContext *avctx, AVPacket *avpkt)
{
    VpeEncCtx *ctx = (VpeEncCtx *)avctx->priv_data;
    AVFrame *frm;
    int stream_size = 0;
    int ret = 0;

    if (ctx->eof == 0) {
        ret = vpe_enc_receive_pic(avctx);
        if (ret){
            if(ret == AVERROR_EXTERNAL)
                av_log(avctx, AV_LOG_ERROR, "vpe_enc_receive_pic failed\n");
            return ret;
        }
    }

    if (!strcmp(avctx->codec->name, "xav1enc_vpe")) {
        stream_size = avctx->width * avctx->height;
    } else {
        ret = vpi_encoder_get_packet_size(ctx->enc_handle, &stream_size);
        if (ret == -1) {
            return AVERROR(EAGAIN);
        } else if (ret == VPI_ENC_EOF) {
            av_log(NULL, AV_LOG_DEBUG, "received EOF from enc\n");
            return AVERROR_EOF;
        } else if (ret != 0) {
            av_log(avctx, AV_LOG_ERROR, "vpi_encoder_get_packet_size, error=%s\n",
                            vpi_error_str(ret));
            return AVERROR_EXTERNAL;
        }
    }

    /*Allocate AVPacket bufffer*/
    ret = av_new_packet(avpkt, stream_size);
    if (ret)
        return ret;

    ctx->vpi_pkt->buffer = avpkt->data;
    ctx->vpi_pkt->size   = stream_size;

    ret = vpi_encoder_get_packet(ctx->enc_handle, (void *)ctx->vpi_pkt);
    if (ret == 0) {
        /*Convert output packet from VpiPacket to AVPacket*/
        vpe_encode_output_packet(ctx->vpi_pkt, avpkt);
    } else {
        av_log(avctx, AV_LOG_ERROR, "enc encode failed, error=%s\n",
               vpi_error_str(ret));
        ret = AVERROR_EXTERNAL;
    }

    frm = vpi_encoder_release_frame(ctx->enc_handle);
    while (frm) {
        /* free the input frame */
        av_frame_free(&frm);
        frm = vpi_encoder_release_frame(ctx->enc_handle);
    }

    return ret;
}

static av_cold int vpe_encode_close(AVCodecContext *avctx)
{
    VpeEncCtx *ctx = (VpeEncCtx *)avctx->priv_data;
    AVFrame *frm;
    vpi_encoder_close(ctx->enc_handle);
    av_frame_free(&ctx->frame);

    if (ctx->input_frm_pool) {
        frm = vpi_encoder_flush_frame(ctx->input_frm_pool);
        while (frm) {
            /* flush the input frame */
            av_frame_free(&frm);
            frm = vpi_encoder_flush_frame(ctx->input_frm_pool);
        }
        vpi_pool_unref(ctx->input_frm_pool);
    }

    if (ctx->init_params) {
        vpi_memory_free_struct(ctx->init_params);
    }

    if (ctx->vpi_pkt) {
        vpi_memory_free_struct(ctx->vpi_pkt);
    }
    if (avctx->extradata)
        av_freep(&avctx->extradata);

    av_buffer_unref(&ctx->hwframe);
    if (ctx->hwdevice)
        av_buffer_unref(&ctx->hwdevice);

    return 0;
}

/* options */
#define OFFSETOPT(x) offsetof(VpeEncCtx, x)

#define FLAGS \
    (AV_OPT_FLAG_ENCODING_PARAM | AV_OPT_FLAG_VIDEO_PARAM | AV_OPT_FLAG_EXPORT)

static const AVOption vpe_encode_options[] = {
    { "crf",
      "VCE Constant rate factor mode. Works with lookahead turned on.",
      OFFSETOPT(crf),
      AV_OPT_TYPE_INT,
      { .i64 = -1 },
      -1,
      51,
      FLAGS },
    { "preset",
      "Set the encoding preset",
      OFFSETOPT(preset),
      AV_OPT_TYPE_STRING,
      { .str = NULL },
      .flags = FLAGS },
    { "profile",
      "Set encode profile. HEVC:0-2; H264:9-12",
      OFFSETOPT(profile),
      AV_OPT_TYPE_STRING,
      { .str = NULL },
      .flags = FLAGS },
    { "level",
      "Set encode level",
      OFFSETOPT(level),
      AV_OPT_TYPE_STRING,
      { .str = NULL },
      .flags = FLAGS },
    { "force8",
      "Force 8 bit output",
      OFFSETOPT(force8),
      AV_OPT_TYPE_INT,
      { .i64 = 0},
      0,
      1,
      .flags = FLAGS },
    { "force_idr",
      "The way to set key frames, can be source/keyframe/interval:xx",
      OFFSETOPT(force_idr),
      AV_OPT_TYPE_STRING,
      { .str = NULL },
      .flags = FLAGS },
    { "enc_params",
      "Override the VPE h264/hevc/av1/xav1 configuration using a :-separated list of "
      "key=value parameters.",
      OFFSETOPT(enc_params),
      AV_OPT_TYPE_STRING,
      { .str = NULL },
      .flags = FLAGS },
    { NULL },
};

static const FFCodecDefault vpe_encode_h264_defaults[] = {
    { NULL },
};

static const FFCodecDefault vpe_encode_hevc_defaults[] = {
    { NULL },
};

static const FFCodecDefault vpe_encode_av1_defaults[] = {
    { NULL },
};

static const FFCodecDefault vpe_encode_jpeg_defaults[] = {
    { NULL },
};

static const AVClass vpe_encode_h264_class = {
    .class_name = "h264e_vpe",
    .item_name  = av_default_item_name,
    .option     = vpe_encode_options,
    .version    = LIBAVUTIL_VERSION_INT,
};

static const AVClass vpe_encode_hevc_class = {
    .class_name = "hevce_vpe",
    .item_name  = av_default_item_name,
    .option     = vpe_encode_options,
    .version    = LIBAVUTIL_VERSION_INT,
};

static const AVClass vpe_encode_av1_class = {
    .class_name = "av1e_vpe",
    .item_name  = av_default_item_name,
    .option     = vpe_encode_options,
    .version    = LIBAVUTIL_VERSION_INT,
};

static const AVClass vpe_encode_jpeg_class = {
    .class_name = "jpege_vpe",
    .item_name  = av_default_item_name,
    .option     = vpe_encode_options,
    .version    = LIBAVUTIL_VERSION_INT,
};

static const AVCodecHWConfigInternal *vpe_enc_hw_configs[] =
    { &(const AVCodecHWConfigInternal){
          .public =
              {
                  .pix_fmt = AV_PIX_FMT_VPE,
                  .methods = AV_CODEC_HW_CONFIG_METHOD_HW_DEVICE_CTX |
                             AV_CODEC_HW_CONFIG_METHOD_HW_FRAMES_CTX,
                  .device_type = AV_HWDEVICE_TYPE_AMA,
              },
          .hwaccel = NULL,
      },
      NULL };

FFCodec ff_h264_vpe_encoder = {
    .p.name           = "h264enc_vpe",
    .p.long_name      = NULL_IF_CONFIG_SMALL("H264 (VPE VC8000E)"),
    .p.type           = AVMEDIA_TYPE_VIDEO,
    .p.id             = AV_CODEC_ID_H264,
    .priv_data_size = sizeof(VpeEncCtx),
    .init           = &vpe_encode_init,
    FF_CODEC_RECEIVE_PACKET_CB(vpe_encode_receive_packet),
    .close          = &vpe_encode_close,
    .p.priv_class     = &vpe_encode_h264_class,
    .p.capabilities   = AV_CODEC_CAP_DELAY | AV_CODEC_CAP_HARDWARE,
    .defaults       = vpe_encode_h264_defaults,
    .p.pix_fmts       = (const enum AVPixelFormat[]) {
        AV_PIX_FMT_VPE,
        AV_PIX_FMT_NONE
  	},
    .hw_configs     = vpe_enc_hw_configs,
    .p.wrapper_name   = "vpe",
};

FFCodec ff_hevc_vpe_encoder = {
    .p.name           = "hevcenc_vpe",
    .p.long_name      = NULL_IF_CONFIG_SMALL("HEVC (VPE VC8000E)"),
    .p.type           = AVMEDIA_TYPE_VIDEO,
    .p.id             = AV_CODEC_ID_HEVC,
    .priv_data_size = sizeof(VpeEncCtx),
    .init           = &vpe_encode_init,
    FF_CODEC_RECEIVE_PACKET_CB(vpe_encode_receive_packet),
    .close          = &vpe_encode_close,
    .p.priv_class     = &vpe_encode_hevc_class,
    .p.capabilities   = AV_CODEC_CAP_DELAY | AV_CODEC_CAP_HARDWARE,
    .defaults       = vpe_encode_hevc_defaults,
    .p.pix_fmts       = (const enum AVPixelFormat[]) {
        AV_PIX_FMT_VPE,
        AV_PIX_FMT_NONE
  	},
    .hw_configs     = vpe_enc_hw_configs,
    .p.wrapper_name   = "vpe",
};

FFCodec ff_av1_vpe_encoder = {
    .p.name           = "av1enc_vpe",
    .p.long_name      = NULL_IF_CONFIG_SMALL("AV1 (VPE VC8000E)"),
    .p.type           = AVMEDIA_TYPE_VIDEO,
    .p.id             = AV_CODEC_ID_AV1,
    .priv_data_size = sizeof(VpeEncCtx),
    .init           = &vpe_encode_init,
    FF_CODEC_RECEIVE_PACKET_CB(vpe_encode_receive_packet),
    .close          = &vpe_encode_close,
    .p.priv_class     = &vpe_encode_av1_class,
    .p.capabilities   = AV_CODEC_CAP_DELAY | AV_CODEC_CAP_HARDWARE,
    .defaults       = vpe_encode_av1_defaults,
    .p.pix_fmts       = (const enum AVPixelFormat[]) {
        AV_PIX_FMT_VPE,
        AV_PIX_FMT_NONE
  	},
    .hw_configs     = vpe_enc_hw_configs,
    .p.wrapper_name   = "vpe",
};

FFCodec ff_jpeg_vpe_encoder = {
    .p.name           = "jpegenc_vpe",
    .p.long_name      = NULL_IF_CONFIG_SMALL("JPEG (VPE VC8000E)"),
    .p.type           = AVMEDIA_TYPE_VIDEO,
    .p.id             = AV_CODEC_ID_MJPEG,
    .priv_data_size = sizeof(VpeEncCtx),
    .init           = &vpe_encode_init,
    FF_CODEC_RECEIVE_PACKET_CB(vpe_encode_receive_packet),
    .close          = &vpe_encode_close,
    .p.priv_class     = &vpe_encode_jpeg_class,
    .p.capabilities   = AV_CODEC_CAP_DELAY | AV_CODEC_CAP_HARDWARE,
    .defaults       = vpe_encode_jpeg_defaults,
    .p.pix_fmts       = (const enum AVPixelFormat[]) {
        AV_PIX_FMT_VPE,
        AV_PIX_FMT_NONE
    },
    .hw_configs     = vpe_enc_hw_configs,
    .p.wrapper_name   = "vpe",
};

FFCodec ff_xav1_vpe_encoder = {
    .p.name           = "xav1enc_vpe",
    .p.long_name      = NULL_IF_CONFIG_SMALL("XILINX AV1 ENC"),
    .p.type           = AVMEDIA_TYPE_VIDEO,
    .p.id             = AV_CODEC_ID_AV1,
    .priv_data_size = sizeof(VpeEncCtx),
    .init           = &vpe_encode_init,
    FF_CODEC_RECEIVE_PACKET_CB(vpe_encode_receive_packet),
    .close          = &vpe_encode_close,
    .p.priv_class     = &vpe_encode_jpeg_class,
    .p.capabilities   = AV_CODEC_CAP_DELAY | AV_CODEC_CAP_HARDWARE,
    .defaults       = vpe_encode_jpeg_defaults,
    .p.pix_fmts       = (const enum AVPixelFormat[]) {
        AV_PIX_FMT_VPE,
        AV_PIX_FMT_NONE
    },
    .hw_configs     = vpe_enc_hw_configs,
    .p.wrapper_name   = "vpe",
};
