/*
* Xilinx ROI side data scale filter
*
* This file is part of FFmpeg.
*
* FFmpeg is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* FFmpeg is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with FFmpeg; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "roi_scale.h"
#include "vpi_types.h"

#include "avfilter.h"
#include "filters.h"
#include "framesync.h"
#include "libavutil/avstring.h"
#include "libavutil/opt.h"
#include "libavutil/pixdesc.h"

#define SBSCALE_MAX_INPUTS 4

#ifndef max
#define max(a,b)            (((a) > (b)) ? (a) : (b))
#endif

//TBD: ROI implementation will eventually update as per VSI new side data framework XLNXS:129

//TBD: roi from file read support to be migrated later as per VSI side data framework
typedef struct SBScaleContext {
    const AVClass *class;
    FFFrameSync fs;
    int nb_inputs;
    int roi_map_type;
    unsigned char *src_roi;
} SBScaleContext;


static int query_formats(AVFilterContext *ctx)
{
    AVFilterFormats *formats = NULL;
    int fmt, ret;

    for (fmt = 0; av_pix_fmt_desc_get(fmt); fmt++) {
        const AVPixFmtDescriptor *desc = av_pix_fmt_desc_get(fmt);
        if (desc->flags & AV_PIX_FMT_FLAG_BITSTREAM)
            continue;
        if (!(desc->flags & AV_PIX_FMT_FLAG_HWACCEL)) {
            // Not usable if there is any subsampling but the format is
            // not planar (e.g. YUYV422).
            if ((desc->log2_chroma_w || desc->log2_chroma_h) &&
                !(desc->flags & AV_PIX_FMT_FLAG_PLANAR))
                continue;
        }
        ret = ff_add_format(&formats, fmt);
        if (ret < 0)
            return ret;
    }

    return ff_set_common_formats(ctx, formats);
}


static av_cold int init(AVFilterContext *ctx)
{
    int ret;
    SBScaleContext *s = NULL;

    s = ctx->priv;

    for (int i = 0; i < s->nb_inputs; i++) {
        AVFilterPad pad = { 0 };

        pad.type = AVMEDIA_TYPE_VIDEO;
        pad.name = av_asprintf("input%d", i);
        if (!pad.name)
            return AVERROR(ENOMEM);

        if ((ret = ff_append_inpad(ctx, &pad)) < 0) {
            av_freep(&pad.name);
            return ret;
        }
    }

    return 0;
}


static av_cold void uninit(AVFilterContext *ctx)
{
    SBScaleContext *s = NULL;

    s = ctx->priv;
    ff_framesync_uninit(&s->fs);

    if (s->src_roi)
        free(s->src_roi);

    for (int i = 0; i < ctx->nb_inputs; i++)
        av_freep(&ctx->input_pads[i].name);
}


static int process_frame(FFFrameSync *fs)
{
    AVFilterContext *ctx = NULL;
    SBScaleContext *s = NULL;
    AVFilterLink *outlink = NULL;
    AVFrame *out, *in[SBSCALE_MAX_INPUTS];
    VpiFrm *roi_pic = NULL;
    VpiFrm *out_pic = NULL;
    VpiChannelInfo *roi_ch_info = NULL;

    ctx = fs->parent;
    s = fs->opaque;
    outlink = ctx->outputs[0];

    for (int i = 0; i < s->nb_inputs; i++)
        ff_framesync_get_frame(&s->fs, i, &in[i], 0);

    out = av_frame_clone(in[0]);
    out_pic = (VpiFrm *)out->data[0];

    for (int i = 1; i < s->nb_inputs; i++) {
        roi_pic = (VpiFrm *)in[i]->data[0];
        if (!roi_pic->vpi_sideband[0] || (roi_pic->vpi_sideband[0] && !roi_pic->vpi_sideband[0]->rc_data)) {
            av_log(ctx, AV_LOG_ERROR, "VpiFrm at input %d does not contain the sideband data.\n", i);
            return AVERROR_EXTERNAL;
        }
        roi_ch_info = &(roi_pic->vpi_sideband[0]->ch_info);
	if (i == 1)
	    memcpy(s->src_roi, roi_pic->vpi_sideband[0]->rc_data, (sizeof(unsigned char)*roi_ch_info->height*roi_ch_info->width));
	else if (i > 1) {
	    unsigned char * additional_roi = (unsigned char *)roi_pic->vpi_sideband[0]->rc_data;
            //TODO: Dimensions comparison for all ROIs after new side data framework.
            for (int i=0; i<(roi_ch_info->height*roi_ch_info->width); i++) {
                s->src_roi[i] = max(s->src_roi[i], additional_roi[i]);
	    }
	}
    }

    roi_pic = (VpiFrm *)in[1]->data[0];
    roi_ch_info = &(roi_pic->vpi_sideband[0]->ch_info);
    if ((in[1]->width == out->width) && (in[1]->height == out->height)) {
        memcpy(roi_pic->vpi_sideband[0]->rc_data, s->src_roi,
                (sizeof(unsigned char)*roi_ch_info->height*roi_ch_info->width));
        out_pic->vpi_sideband[0] = roi_pic->vpi_sideband[0];
    }
    else {
        out_pic->vpi_sideband[0] = roi_pic->vpi_sideband[0];
        out_pic->vpi_sideband[0]->ch_info.width = out->width/16;
        out_pic->vpi_sideband[0]->ch_info.height = out->height/16;
        scale_roi(s->src_roi, out_pic->vpi_sideband[0]->rc_data, in[1]->width, in[1]->height, in[0]->width, in[0]->height);
    }

    if (!s->roi_map_type) {
        unsigned char * map;
        int roi_size;
        map = (unsigned char *)(out_pic->vpi_sideband[0]->rc_data);
        roi_size = out_pic->vpi_sideband[0]->ch_info.width*out_pic->vpi_sideband[0]->ch_info.height;
        for (int i = 0; i < roi_size; ++i) {
            map[i] = -((max(0, (map[i]) - 128) * 8 + 32) >> 6);
        }
    }

    return ff_filter_frame(outlink, out);
}


static int config_output(AVFilterLink *link)
{
    SBScaleContext *s = NULL;
    AVFilterContext *ctx = NULL;
    FFFrameSyncIn *in = NULL;
    int ret;
    int roi_width_aligned = 0;
    int roi_height_aligned = 0;

    s = link->src->priv;
    s->fs.on_event = process_frame;
    s->fs.opaque   = s;
    ctx = link->src;

    roi_width_aligned = (ctx->inputs[1]->w + 15) >> 4;
    roi_height_aligned = (ctx->inputs[1]->h + 15) >> 4;
    s->src_roi = (unsigned char *)malloc(sizeof(unsigned char) * roi_width_aligned * roi_height_aligned);

    ret = ff_framesync_init(&s->fs, ctx, s->nb_inputs);
    if (ret < 0)
        return ret;

    in = s->fs.in;
    for (int i = 0; i < s->nb_inputs; i++) {
        in[i].time_base = ctx->inputs[i]->time_base;
	in[i].sync   = 1;
	in[i].before = EXT_STOP;
	in[i].after  = EXT_INFINITY;
    }

    ret = ff_framesync_configure(&s->fs);
    link->time_base = s->fs.time_base;

    return ret;
}


static int activate(AVFilterContext *ctx)
{
    SBScaleContext *s = ctx->priv;
    return ff_framesync_activate(&s->fs);
}


#define OFFSET(x) offsetof(SBScaleContext, x)
#define FLAGS AV_OPT_FLAG_FILTERING_PARAM|AV_OPT_FLAG_VIDEO_PARAM

static const AVOption vip_sbscale_options[] = {
    { "inputs", "set number of inputs", OFFSET(nb_inputs), AV_OPT_TYPE_INT, {.i64=2},  2, SBSCALE_MAX_INPUTS, .flags = FLAGS },
    { "roi_map_type", "type of roi map", OFFSET(roi_map_type), AV_OPT_TYPE_INT, {.i64=0},  0, 1, .flags = FLAGS },
    { NULL }
};

AVFILTER_DEFINE_CLASS(vip_sbscale);

static const AVFilterPad avfilter_vf_sbscale_outputs[] = {
    {
        .name         = "default",
        .type         = AVMEDIA_TYPE_VIDEO,
        .config_props = config_output,
    },
    { NULL }
};

const AVFilter ff_vf_vip_sbscale = {
    .name            = "roi_sbscale_ama",
    .description     = NULL_IF_CONFIG_SMALL("Scale the roi sb and move it on frame to be encoded."),
    .priv_size       = sizeof(SBScaleContext),
    .priv_class      = &vip_sbscale_class,
    .init            = init,
    .uninit          = uninit,
    .activate        = activate,
    .outputs         = avfilter_vf_sbscale_outputs,
    .nb_outputs      = 1,
    FILTER_QUERY_FUNC(query_formats),
    .flags           = AVFILTER_FLAG_SUPPORT_TIMELINE_INTERNAL,
};
